/* $Id: graphics.c,v 1.6 2006/01/01 11:01:03 cegger Exp $
***************************************************************************

   X server for LibGGI - Handling of graphics output

   Copyright (C) 1997	   Jason McMullan	[jmcc@cs.cmu.edu]
   Copyright (C) 1997	   Michael Krause	[rawstyle@ms.demo.org]
   Copyright (C) 1998-2000 Marcus Sundberg	[marcus@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************
*/

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "X11/Xos.h"

#define PSZ 8
#include "cfb.h"
#include "cfb16.h"
#include "cfb24.h"
#include "cfb32.h"
#include "scrnintstr.h"
#include "servermd.h"
#include "colormapst.h"
#include "mipointer.h"

#include "xggi.h"
#include "xggi_misc.h"
#include "xggi_Xext.h"


#define MAX_COLORS 256 /* For the PseudoColor modes.. */


/* Common pixmap formats (never more than MAXFORMATS == 8) */

static PixmapFormatRec pixmap_formats[] = {
	{ 1,	1,	BITMAP_SCANLINE_PAD },
#if 0
/* Not supported by mfb/cfb/mi sw renderers. Thus we can't use it... :( */
	{ 2,	8,	BITMAP_SCANLINE_PAD },
	{ 4,	8,	BITMAP_SCANLINE_PAD },
#endif
	{ 8,	8,	BITMAP_SCANLINE_PAD },
	{ 15,	16,	BITMAP_SCANLINE_PAD },
	{ 16,	16,	BITMAP_SCANLINE_PAD },
	{ 24,	32,	BITMAP_SCANLINE_PAD },
#ifdef RENDER
	{ 32,	32,	BITMAP_SCANLINE_PAD },
#endif
};

#define pixmap_numFormats	(sizeof(pixmap_formats)/sizeof(PixmapFormatRec))



struct xggiGlobalInfo xggiInfo = {
	0,	/* Number of screens */
	1, 1,	/* Screen rows and columns */
	1,	/* Allow zapping ? */
	1,	/* Automaticly set GII_CTRLALT_VTSWITCH */
	0,	/* Don't open new VT by default. */
	0,	/* XGGIGeneration */
};


struct xggiScreenInfo *xggiScreens;


static char *targetlist = NULL, *modelist = NULL;
static int defaultbpp = 0;
static int ggiinit = 0;

#if 0
void DPMSSet(int level);
void DPMSSet(int level)
{
}
#endif

static void _ggiexit(void)
{
	if (ggiinit) {
		xggiScreenFree();

		ggiExit();
		ggiinit = 0;
	}
}
	
void AbortDDX(void)
{
	_ggiexit();
}

void ddxGiveUp(void)
{
	_ggiexit();
}


/* ddxInitGlobals - called by |InitGlobals| from os/util.c */
void ddxInitGlobals(void)
{
}


void OsVendorInit(void)
{
}


void OsVendorFatalError(void)
{
	/* don't call _ggiexit() here. */
}


void ddxUseMsg(void)
{
	ErrorF(
"XGGI adds the following arguments:\n"
"-targets targetlist    the target(s) we should open the screen(s) on\n"
"-modes ggimodelist     the mode(s) we should use\n"
"-bpp bitdepth          XFree86 compability option\n"
"-rows numrows          number of screen rows in multihead configurations\n"
#if 0 /* Rows are enough to describe the layout */
"-columns               number of screen columns in multihead configurations\n"
#endif
"-buttons nr-buttons    specify number of pointer buttons\n"
"-dontzap               disable the <Crtl><Alt><BS> server abort sequence\n"
"-ggivtswitch           don't set GII_CTRLALT_VTSWITCH automaticly\n"
"-newvt                 set GGI_NEWVT environment variable\n"
"-noxfreeemu            provided only for backward compability\n");
}


int ddxProcessArgument(int argc, char *argv[], int i)
{
	if (!strcmp(argv[i], "-targets")) {
		if (i + 1 >= argc) {
			ErrorF("Option `-targets' requires an argument\n");
			return 1;
		}
		targetlist = safe_alloc(strlen(argv[i+1])+1);
		strcpy(targetlist, argv[i+1]);
		return 2;
	}
	if (!strcmp(argv[i], "-modes")) {
		if (i + 1 >= argc) {
			ErrorF("Option `-modes' requires an argument\n");
			return 1;
		}
		modelist = safe_alloc(strlen(argv[i+1])+1);
		strcpy(modelist, argv[i+1]);
		return 2;
	}
	if (!strcmp(argv[i], "-bpp")) {
		if (i + 1 >= argc) {
			ErrorF("Option `-bpp' requires an argument\n");
			return 1;
		}
		defaultbpp = atoi(argv[i+1]);
		return 2;
	}
	if (!strcmp(argv[i], "-rows")) {
		if (i + 1 >= argc) {
			ErrorF("Option `-rows' requires an argument\n");
			return 1;
		}
		xggiInfo.rows = atoi(argv[i+1]);
		return 2;
	}
#if 0
	if (!strcmp(argv[i], "-columns")) {
		if (i + 1 >= argc) {
			ErrorF("Option `-columns' requires an argument\n");
			return 1;
		}
		xggiInfo.columns = atoi(argv[i+1]);
		return 2;
	}
#endif
	if (!strcmp(argv[i], "-buttons")) {
		if (i + 1 >= argc) {
			ErrorF("Option `-buttons' requires an argument\n");
			return 1;
		}
		MASTER_SCREEN.ptrbuttons = atoi(argv[i+1]);
		return 2;
	}
	if (!strcmp(argv[i], "-dontzap")) {
		xggiInfo.allowzap = 0;
		return 1;
	}
	if (!strcmp(argv[i], "-ggivtswitch")) {
		xggiInfo.ctrlalt_vtswitch = 0;
		return 1;
	}
	if (!strcmp(argv[i], "-newvt")) {
		xggiInfo.newvt = 1;
		return 1;
	}
	if (!strcmp(argv[i], "-noxfreeemu")) {
		return 1;
	}

	return 0;
}

#ifdef DDXTIME /* from ServerOSDefines */
CARD32 GetTimeInMillis(void)
{
	struct timeval  tp;

	ggCurTime(&tp);
	return TVAL2TIME(tp);
}
#endif


static ColormapPtr InstalledMaps[MAXSCREENS];

static int xggiListInstalledColormaps(ScreenPtr pScreen, Colormap *pmaps)
{
	/* By the time we are processing requests, we can guarantee that there
	 * is always a colormap installed
	 */
	*pmaps = InstalledMaps[pScreen->myNum]->mid;
	return (1);
}


static void xggiInstallColormap(ColormapPtr pmap)
{
	int idx = pmap->pScreen->myNum;
	ColormapPtr oldpmap = InstalledMaps[idx];

	if (pmap != oldpmap) {
		int entries;
		VisualPtr pVisual;
		Pixel *     ppix;
		xrgb *      prgb;
		xColorItem *defs;
		int i;

		if (oldpmap != (ColormapPtr)None)
			WalkTree(pmap->pScreen, TellLostMap, (char *)&oldpmap->mid);
		/* Install pmap */
		InstalledMaps[idx] = pmap;
		WalkTree(pmap->pScreen, TellGainedMap, (char *)&pmap->mid);

		entries = pmap->pVisual->ColormapEntries;
		pVisual = pmap->pVisual;

		ppix = (Pixel *)ALLOCATE_LOCAL(entries * sizeof(Pixel));
		prgb = (xrgb *)ALLOCATE_LOCAL(entries * sizeof(xrgb));
		defs = (xColorItem *)ALLOCATE_LOCAL(entries * sizeof(xColorItem));

		for (i = 0; i < entries; i++)  ppix[i] = i;
		/* XXX truecolor */
		QueryColors(pmap, entries, ppix, prgb);

		for (i = 0; i < entries; i++) { /* convert xrgbs to xColorItems */
			defs[i].pixel = ppix[i] & 0xff; /* change pixel to index */
			defs[i].red = prgb[i].red;
			defs[i].green = prgb[i].green;
			defs[i].blue = prgb[i].blue;
			defs[i].flags =  DoRed|DoGreen|DoBlue;
		}
		(*pmap->pScreen->StoreColors)(pmap, entries, defs);

		DEALLOCATE_LOCAL(ppix);
		DEALLOCATE_LOCAL(prgb);
		DEALLOCATE_LOCAL(defs);
	}
}

static void xggiUninstallColormap(ColormapPtr pmap)
{
	ColormapPtr curpmap = InstalledMaps[pmap->pScreen->myNum];

	if (pmap == curpmap) {
		if ((unsigned)pmap->mid != pmap->pScreen->defColormap) {
			curpmap = (ColormapPtr)
				LookupIDByType(pmap->pScreen->defColormap,
					       RT_COLORMAP);
			(*pmap->pScreen->InstallColormap)(curpmap);
		}
	}
}

static void xggiStoreColors(ColormapPtr pmap, int ndef, xColorItem  *pdefs)
{
	int i;
	ggi_color cmap[MAX_COLORS];
	int scr;

	if (pmap != InstalledMaps[pmap->pScreen->myNum]) {
		return;
	}

	if ((pmap->pVisual->class | DynamicClass) == DirectColor) {
		return;
	}

	scr = xggiGetScreenIdx(pmap->pScreen);

	ggiGetPalette(MASTER_VISUAL_VIS(scr), 0, MAX_COLORS, cmap);

	for (i = 0; i < ndef; i++) {
		if (pdefs[i].pixel >= MAX_COLORS) continue;

		if (pdefs[i].flags & DoRed) {
			cmap[pdefs[i].pixel].r = pdefs[i].red;
		}
		if (pdefs[i].flags & DoGreen) {
			cmap[pdefs[i].pixel].g = pdefs[i].green;
		}
		if (pdefs[i].flags & DoBlue) {
			cmap[pdefs[i].pixel].b = pdefs[i].blue;
		}
	}

	ggiSetPalette(MASTER_VISUAL_VIS(scr), 0, MAX_COLORS, cmap);
}

static Bool xggiSaveScreen(ScreenPtr pScreen, int on)
{
#if 0
	switch (what) {
	case SCREEN_SAVER_ON:
		break;
      
	case SCREEN_SAVER_OFF:
		break;
      
	case SCREEN_SAVER_FORCER:
		break;
      
	case SCREEN_SAVER_CYCLE:
		break;
	}
#endif

	return TRUE;
}


#if 0	/* Defined, but unused */
static void do_blit(ScreenPtr pScreen, int bits_per_pixel,
			DrawablePtr src, DrawablePtr dest)
{
	DDXPointRec pixPt;
	BoxRec    pixBox;
	RegionRec pixReg;

	pixBox.x1 = 0;
	pixBox.x2 = pScreen->width;
	pixBox.y1 = 0;
	pixBox.y2 = pScreen->height;
	pixPt.x = pixPt.y = 0;
	/*pScreen->RegionInit(&pixReg, &pixBox, 1);*/

	if (bits_per_pixel == 1) {
		mfbDoBitblt(src, dest, GXcopy, &pixReg, &pixPt);
	} else if (bits_per_pixel == 8) {
		cfbDoBitblt(src, dest, GXcopy, &pixReg, &pixPt, 0xFF);
	} else if (bits_per_pixel == 16) {
		cfb16DoBitblt(src, dest, GXcopy, &pixReg, &pixPt, 0xFFFF);
	} else if (bits_per_pixel == 24) {
		cfb24DoBitblt(src, dest, GXcopy, &pixReg, &pixPt, 0xFFFFFF);
	} else if (bits_per_pixel == 32) {
		cfb32DoBitblt(src, dest, GXcopy, &pixReg, &pixPt, 0xFFFFFFFF);
	}
}
#endif


static void dounmap(int idx)
{
#if 0
	ScreenPtr pScreen = xggiScreens[idx].screen;
	int bits_per_pixel = xggiScreens[idx].bitsPerPixel;
	PixmapPtr frontbuf;

	xggiScreens[idx].backbuffer
		= pScreen->CreatePixmap(pScreen, pScreen->width,
					pScreen->height, pScreen->rootDepth);
	if (!xggiScreens[idx].backbuffer) {
		FatalError("Unable to allocate backbuffer!\n");
	}

	if (bits_per_pixel == 16) {
		frontbuf = (PixmapPtr)pScreen->devPrivates[cfbScreenPrivateIndex].ptr;
	} else if (bits_per_pixel == 24) {
		frontbuf = (PixmapPtr)pScreen->devPrivates[cfbScreenPrivateIndex].ptr;
	} else if (bits_per_pixel == 32) {
		frontbuf = (PixmapPtr)pScreen->devPrivates[cfbScreenPrivateIndex].ptr;
	} else {
		frontbuf = (PixmapPtr)pScreen->devPrivate;
	}

	do_blit(pScreen, bits_per_pixel,
		&frontbuf->drawable,
		&xggiScreens[idx].backbuffer->drawable);

	frontbuf->devPrivate.ptr = xggiScreens[idx].backbuffer->devPrivate.ptr;
	xggiScreens[idx].ismapped = 0;
#endif
}

	
static void domap(int idx)
{
#if 0
	ScreenPtr pScreen = xggiScreens[idx].screen;
	int bits_per_pixel = xggiScreens[idx].bitsPerPixel;
	PixmapPtr frontbuf;

	if (bits_per_pixel == 16) {
		frontbuf = (PixmapPtr)pScreen->devPrivates[cfbScreenPrivateIndex].ptr;
	} else if (bits_per_pixel == 24) {
		frontbuf = (PixmapPtr)pScreen->devPrivates[cfbScreenPrivateIndex].ptr;
	} else if (bits_per_pixel == 32) {
		frontbuf = (PixmapPtr)pScreen->devPrivates[cfbScreenPrivateIndex].ptr;
	} else {
		frontbuf = (PixmapPtr)pScreen->devPrivate;
	}

	frontbuf->devPrivate.ptr = xggiScreens[idx].pfbMemory;

	do_blit(pScreen, bits_per_pixel,
		&xggiScreens[idx].backbuffer->drawable,
		&frontbuf->drawable);
	
	pScreen->DestroyPixmap(xggiScreens[idx].backbuffer);
	xggiScreens[idx].backbuffer = NULL;
	xggiScreens[idx].ismapped = 1;
#endif
}


void xggiUnmapDisplay(void)
{
	int i;

	for (i = 0; i < xggiInfo.numscreens; i++) {
		dounmap(i);
	}
}


void xggiMapDisplay(void)
{
	int i;

	for (i = 0; i < xggiInfo.numscreens; i++) {
		domap(i);
	}
}


#define XYtoI(x,y)	((y) * xggiInfo.columns + (x))
#define ItoXY(x,y,i) \
do { \
	(x) = (i) % xggiInfo.columns; \
	(y) = (i) / xggiInfo.columns; \
}while(0)

static Bool xggiCursorOffScreen(ScreenPtr *ppScreen, int *x, int *y)
{
	int scr = xggiGetScreenIdx(*ppScreen);
	int maxscr = xggiInfo.numscreens-1;
	int newscr;
	int xs, ys;
	
	if (maxscr < 1) return FALSE;

	if (*x >= MASTER_VISUAL_MODE(scr).visible.x) {
		ItoXY(xs, ys, scr);
		xs++;
		if (xs >= xggiInfo.columns) return FALSE;
		newscr = XYtoI(xs, ys);
		if (newscr > maxscr) return FALSE;

		*ppScreen = SCREEN_SCR(newscr);
		*x = 0;
		return TRUE;
	} else if (*x < 0) {
		ItoXY(xs, ys, scr);
		xs--;
		if (xs < 0) return FALSE;
		newscr = XYtoI(xs, ys);
		if (newscr < 0) return FALSE;

		*ppScreen = SCREEN_SCR(newscr);
		*x = MASTER_VISUAL_MODE(newscr).visible.x - 1;
		return TRUE;
	} else if (*y >= MASTER_VISUAL_MODE(scr).visible.y) {
		ItoXY(xs, ys, scr);
		ys++;
		if (ys >= xggiInfo.rows) return FALSE;
		newscr = XYtoI(xs, ys);
		if (newscr > maxscr) return FALSE;

		*ppScreen = SCREEN_SCR(newscr);
		*y = 0;
		return TRUE;
	} else if (*y < 0) {
		ItoXY(xs, ys, scr);
		ys--;
		if (ys < 0) return FALSE;
		newscr = XYtoI(xs, ys);
		if (newscr < 0) return FALSE;

		*ppScreen = SCREEN_SCR(newscr);
		*y = MASTER_VISUAL_MODE(newscr).visible.y - 1;
		return TRUE;
	}

	return FALSE;
}

static void
xggiCrossScreen(ScreenPtr pScreen, Bool entering)
{
}

static miPointerScreenFuncRec xggiPointerCursorFuncs =
{
	xggiCursorOffScreen,
	xggiCrossScreen,
	miPointerWarpCursor
};

static Bool
xggiFBInitProc(int idx, ScreenPtr pScreen,
	       int argc, char *argv[])
{
	const int dpix = 100, dpiy = 100;
	int scr = idx;
	Bool ret;

#ifdef XGGI_DEBUG
	ErrorF("xggiFBInitProc: Index: %d, pScreen %p, depth: %d\n", idx, pScreen,
	       MASTER_VISUAL_DEPTH(scr));
#endif

	SCREEN_SCR(scr) = pScreen;

	ret = xggiScreenInit(pScreen, MASTER_VISUAL(scr).pfbMemory,
			     MASTER_VISUAL_WIDTH(scr), MASTER_VISUAL_HEIGHT(scr),
			     dpix, dpiy, MASTER_VISUAL_STRIDE(scr));
	if (!ret) {
		ErrorF("xggiScreenInit failed\n");
		return FALSE;
	}

	pScreen->SaveScreen = xggiSaveScreen;

	switch (MASTER_VISUAL_DEPTH(scr)) {
	case 1:
		pScreen->InstallColormap = mfbInstallColormap;
		pScreen->UninstallColormap = mfbUninstallColormap;
		pScreen->ListInstalledColormaps = mfbListInstalledColormaps;
		pScreen->StoreColors = (void *)NoopDDA;
		break;
	case 15:
	case 16:
	case 24:
	case 32:
		pScreen->InstallColormap = cfbInstallColormap;
		pScreen->UninstallColormap = cfbUninstallColormap;
		pScreen->ListInstalledColormaps = cfbListInstalledColormaps;
		pScreen->StoreColors = (void *)NoopDDA;
		break;
	default:
		pScreen->InstallColormap = xggiInstallColormap;
		pScreen->UninstallColormap = xggiUninstallColormap;
		pScreen->ListInstalledColormaps = xggiListInstalledColormaps;
		pScreen->StoreColors = xggiStoreColors;
		break;
	}

	miDCInitialize(pScreen, &xggiPointerCursorFuncs);

	if (MASTER_VISUAL_DEPTH(scr) == 1) {
		ret = mfbCreateDefColormap(pScreen);
	} else {
		ret = cfbCreateDefColormap(pScreen);
	}
	if (!ret) ErrorF("[c|m]fbCreateDefColormap failed\n");

	return ret;
}


static int count_targets(const char *targets)
{
	char tgttemp[2048];
	int count = 0;

	while (1) {
		targets = ggParseTarget(targets, tgttemp, 2048);

		if (targets == NULL) break;

		count++;

		while (*targets && isspace((uint8_t)*targets)) targets++;
		if (*targets == '\0') {
			break;
		}

		if (*targets != ':') {
			FatalError("Missing ':' between targets: %c\n", *targets);
		}

		targets++;  /* skip ':' */
	}

	return count;
}	


static ggi_graphtype bpp2gt(int bpp)
{
	switch (bpp) {
	case 1:		return GT_1BIT;
	case 2:		return GT_2BIT;
	case 4:		return GT_4BIT;
	case 8:		return GT_8BIT;
	case 15:	return GT_15BIT;
	case 16:	return GT_16BIT;
	case 24:	return GT_24BIT;
	case 32:	return GT_32BIT;
	}	/* switch */
        
	return GT_AUTO;
}	/* bpp2gt */


static void
InitGGI(ScreenInfo *_screenInfo, int argc, char *argv[])
{
	int i;
	const char *targets = targetlist;
	const char *modes = modelist;

	if (targets != NULL) {
		xggiInfo.numscreens = count_targets(targets);
		if (xggiInfo.rows < 1) {
			xggiInfo.rows = 1;
			xggiInfo.columns = xggiInfo.numscreens;
		} else if (xggiInfo.rows > xggiInfo.numscreens) {
			xggiInfo.rows = xggiInfo.numscreens;
			xggiInfo.columns = 1;
		} else {
			xggiInfo.columns = xggiInfo.numscreens / xggiInfo.rows;
			if (xggiInfo.numscreens % xggiInfo.rows) {
				xggiInfo.columns++;
			}
		}
	} else {
		/* One screen on default LibGGI visual */
		xggiInfo.numscreens = xggiInfo.rows = xggiInfo.columns = 1;
	}


	xggiScreenAlloc(1 /* numvisuals */);


	if (xggiInfo.ctrlalt_vtswitch) {
		putenv("GII_CTRLALT_VTSWITCH=1");
	}
	if (xggiInfo.newvt) {
		putenv("GGI_NEWVT=1");
	}

	if (ggiInit() < 0) {
		FatalError("Unable to init LibGGI!\n");
	}
	ggiinit = 1;

	for (i = 0; i < xggiInfo.numscreens; i++) {
		char tgttemp[2048];
		char *currtarget = NULL;
		int ret;

		if (targets) {
			targets = ggParseTarget(targets, tgttemp, 2048);

			if (*tgttemp != '\0') {
				currtarget = tgttemp;
			}

			while (*targets && isspace((uint8_t)*targets)) targets++;
			targets++;  /* skip ':' */
		}

		MASTER_VISUAL_VIS(i) = ggiOpen(currtarget);
		if (MASTER_VISUAL_VIS(i) == NULL) {
			FatalError("Unable to open LibGGI visual: \"%s\"\n",
				   currtarget ? currtarget : "(NULL)");
		}

#if 0	/* Can we hook in ggiFlush() somewhere? */
		ggiSetFlags(MASTER_VISUAL_VIS(i), GGIFLAG_ASYNC);
#endif

		if (modes != NULL &&
		    (modes = ggParseTarget(modes, tgttemp, 2048)) != NULL)
		{
			ggiParseMode(tgttemp, &MASTER_VISUAL_MODE(i));

			while (*modes && isspace((int)*modes)) modes++;
			if (*modes == '\0') {
				modes = NULL;
				goto end_of_modeparsing;
			}
			if (*modes != ':') {
				FatalError("Missing ':' between modes\n");
			}

			modes++;  /* skip ':' */
		} else {
			ggiParseMode("", &MASTER_VISUAL_MODE(i));
		}
	  end_of_modeparsing:

		if (MASTER_VISUAL_MODE(i).graphtype == GT_AUTO
		   && defaultbpp != 0)
		{
			MASTER_VISUAL_MODE(i).graphtype = bpp2gt(defaultbpp);
		}	/* if */

		ret = xggiSwitchMode(i, &MASTER_VISUAL_MODE(i));
		if (ret < 0) {
			switch (ret) {
			case -30:
				FatalError("All GGI visuals must be of the same depth.\n");
			case -31:
				FatalError("LibGGI can not set any modes at all!\n");
			case -32:
				FatalError("This mode and target has no suitable DirectBuffer\n");
			case -33:
				FatalError("Unable to acquire DirectBuffer\n");
			}	/* switch */
		}	/* if */
	}
}


void InitOutput(ScreenInfo *_screenInfo, int argc, char *argv[])
{
	int i;

	if (!ggiinit) {
		InitGGI(_screenInfo, argc, argv);
	}

	/* Enable BackingStore by default */
	enableBackingStore = TRUE;

	/* Set up pixmap formats */
	memcpy(_screenInfo->formats, pixmap_formats,
		sizeof(pixmap_formats));
	_screenInfo->numPixmapFormats = pixmap_numFormats;


	_screenInfo->imageByteOrder = IMAGE_BYTE_ORDER;
	_screenInfo->bitmapScanlineUnit = BITMAP_SCANLINE_UNIT;
	_screenInfo->bitmapScanlinePad = BITMAP_SCANLINE_PAD;
	_screenInfo->bitmapBitOrder = BITMAP_BIT_ORDER;

	/* Add screens */
	for (i = 0; i < xggiInfo.numscreens; i++) {
		if (AddScreen(xggiFBInitProc, argc, argv) == -1) {
			FatalError("Couldn't add screen");
		}	/* if */

#ifdef RENDER
		xggiXextRenderInit(i);
#endif
#ifdef RANDR
		xggiXextRandRInit(i);
#endif
	}	/* for */


}	/* InitOutput */
